use proc_macro::TokenStream;
use quote::quote;
use syn::{parse_macro_input, DeriveInput, Expr, Lit, Meta};

#[proc_macro_derive(FromChar, attributes(value))]
pub fn as_char_derive(input: TokenStream) -> TokenStream {
    let input = parse_macro_input!(input as DeriveInput);
    let name = input.ident;

    let variants = match input.data {
        syn::Data::Enum(data) => data.variants,
        _ => panic!("FromChar can only be derived for enums"),
    };

    let variant_matches = variants.iter().map(|variant| {
        let variant_name = &variant.ident;
        let char_attr = variant
            .attrs
            .iter()
            .find(|attr| attr.path().is_ident("value"))
            .expect("Each variant must have a #[value = \"...\"] attribute");

        let char_value = match &char_attr.meta {
            Meta::NameValue(meta) => match &meta.value {
                Expr::Lit(expr) => match &expr.lit {
                    Lit::Str(str) => str.value(),
                    Lit::Byte(byte) => byte.value().to_string(),
                    Lit::Char(char) => char.value().to_string(),
                    _ => panic!("#[value] attribute must have a string literal value"),
                },
                _ => panic!("#[value] attribute must have a string literal value"),
            },
            _ => panic!("Invalid #[value] attribute format"),
        };

        let char_literal = syn::LitChar::new(
            char_value.chars().next().unwrap(),
            proc_macro2::Span::call_site(),
        );

        (variant_name, char_literal)
    });

    let from_char_variants = variant_matches.clone().map(|(variant, ch)| {
        quote! { #ch => Ok(#name::#variant), }
    });

    let to_char_variants = variant_matches.clone().map(|(variant, ch)| {
        quote! { #name::#variant => #ch, }
    });

    let chars = variant_matches.map(|(_, ch)| quote! { #ch, });

    let expanded = quote! {
        impl char_enum::FromChar for #name {
            type Err = char_enum::FromCharError;

            fn from_char(c: char) -> Result<Self, Self::Err> {
                match c {
                    #(#from_char_variants)*
                    ch => Err(char_enum::FromCharError(ch))
                }
            }
        }

        impl std::str::FromStr for #name {
            type Err = char_enum::FromStrError;

            fn from_str(s: &str) -> Result<Self, Self::Err> {
                match s.chars().next() {
                    Some(c) => Ok(Self::from_char(c)?),
                    None => Err(char_enum::FromStrError(s.to_string())),
                }
            }
        }

        impl char_enum::ToChar for #name {
            fn to_char(&self) -> char {
                match self {
                    #(#to_char_variants)*
                }
            }
        }

        impl std::fmt::Display for #name {
            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                write!(f, "{}", self.to_char())
            }
        }

        impl char_enum::AsIter<char> for #name {
            fn iter() -> impl Iterator<Item = char> {
                [#(#chars)*].into_iter()
            }
        }
    };

    TokenStream::from(expanded)
}
