use itertools::Itertools;
use std::{
    convert::Infallible,
    error::Error,
    ffi::{OsStr, OsString},
    fmt::Display,
    ops::{Deref, DerefMut},
    path::PathBuf,
    process::Command,
    str::FromStr,
};

use crate::utils::DisplayString;

use super::{
    id_mapping::{ParseUntypedIdRangeError, UntypedIdRange},
    permission::Permissions,
    FileSystem,
};

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct ParseKeyValuePairError;

impl Display for ParseKeyValuePairError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "could not parse key value pair")
    }
}

impl Error for ParseKeyValuePairError {}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct KeyValuePair {
    name: DisplayString,
    value: Option<DisplayString>,
}

impl KeyValuePair {
    pub fn new(name: impl Into<DisplayString>, value: Option<impl Into<DisplayString>>) -> Self {
        Self {
            name: name.into(),
            value: value.map(Into::into),
        }
    }

    pub fn into_tuple(self) -> (DisplayString, Option<DisplayString>) {
        (self.name, self.value)
    }
}

impl From<KeyValuePair> for (DisplayString, Option<DisplayString>) {
    fn from(kvp: KeyValuePair) -> Self {
        kvp.into_tuple()
    }
}

impl FromStr for KeyValuePair {
    type Err = ParseKeyValuePairError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.split_once('=') {
            Some((name, value)) => Ok(Self::new(name, Some(value))),
            None => Ok(Self::new(s, None::<&str>)),
        }
    }
}

impl Display for KeyValuePair {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match &self.value {
            Some(value) => write!(f, "{}={}", self.name, value),
            None => write!(f, "{}", self.name),
        }
    }
}

#[derive(Debug, PartialEq, Eq)]
pub struct ParseAccessError;

impl Display for ParseAccessError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "received invalid access type")
    }
}

impl Error for ParseAccessError {}

#[derive(Clone, Copy, Debug, Default, PartialEq, Eq)]
pub enum Access {
    ReadOnly,
    #[default]
    ReadWrite,
}

impl FromStr for Access {
    type Err = ParseAccessError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "read-only" => Ok(Self::ReadOnly),
            "read-write" => Ok(Self::ReadWrite),
            _ => Err(ParseAccessError),
        }
    }
}

impl Display for Access {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Access::ReadOnly => "read-only",
                Access::ReadWrite => "read-write",
            }
        )
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct ParseDeviceIdTypeError(String);

impl Display for ParseDeviceIdTypeError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{} is not a valid device type", self.0)
    }
}

impl Error for ParseDeviceIdTypeError {}

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum DeviceId {
    Label(DisplayString),
    Uuid(DisplayString),
    PartLabel(DisplayString),
    PartUuid(DisplayString),
    Id(DisplayString),
    Other(DisplayString),
}

impl FromStr for DeviceId {
    type Err = ParseDeviceIdTypeError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.split_once('=') {
            Some((key, value)) => match key {
                "LABEL" => Ok(Self::Label(value.into())),
                "UUID" => Ok(Self::Uuid(value.into())),
                "PARTLABEL" => Ok(Self::PartLabel(value.into())),
                "PARTUUID" => Ok(Self::PartUuid(value.into())),
                "ID" => Ok(Self::Id(value.into())),
                key => Err(ParseDeviceIdTypeError(key.to_string())),
            },
            None => Ok(Self::Other(s.into())),
        }
    }
}

impl Display for DeviceId {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                DeviceId::Label(label) => format!("LABEL={label}"),
                DeviceId::Uuid(uuid) => format!("UUID={uuid}"),
                DeviceId::PartLabel(label) => format!("PARTLABEL={label}"),
                DeviceId::PartUuid(uuid) => format!("PARTUUID={uuid}"),
                DeviceId::Id(id) => format!("ID={id}"),
                DeviceId::Other(other) => other.to_string(),
            }
        )
    }
}

impl From<String> for DeviceId {
    fn from(value: String) -> Self {
        Self::from_str(&value).unwrap_or_else(|_v| Self::Other(value.into()))
    }
}

impl From<&str> for DeviceId {
    fn from(value: &str) -> Self {
        Self::from_str(value).unwrap_or_else(|_v| Self::Other(value.into()))
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseOptionsModeError;

impl Display for ParseOptionsModeError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "could not parse options mode")
    }
}

impl Error for ParseOptionsModeError {}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum OptionsMode {
    Ignore,
    Append,
    Prepend,
    Replace,
}

impl FromStr for OptionsMode {
    type Err = ParseOptionsModeError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.trim() {
            "ignore" => Ok(Self::Ignore),
            "append" => Ok(Self::Append),
            "prepend" => Ok(Self::Prepend),
            "replace" => Ok(Self::Replace),
            _ => Err(ParseOptionsModeError),
        }
    }
}

impl Display for OptionsMode {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                OptionsMode::Ignore => "ignore",
                OptionsMode::Append => "append",
                OptionsMode::Prepend => "prepend",
                OptionsMode::Replace => "replace",
            }
        )
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseOptionsSourceError(String);

impl Display for ParseOptionsSourceError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "received an invalid option source: {}", self.0)
    }
}

impl Error for ParseOptionsSourceError {}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum OptionsSource {
    Mtab,
    Fstab,
    Disable,
}

impl FromStr for OptionsSource {
    type Err = ParseOptionsSourceError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "mtab" => Ok(Self::Mtab),
            "fstab" => Ok(Self::Fstab),
            "disable" => Ok(Self::Disable),
            e => Err(ParseOptionsSourceError(e.to_string())),
        }
    }
}

impl Display for OptionsSource {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                OptionsSource::Mtab => "mtab",
                OptionsSource::Fstab => "fstab",
                OptionsSource::Disable => "disable",
            }
        )
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseUserMapError;

impl Display for ParseUserMapError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "could not parse user-map arguments")
    }
}

impl Error for ParseUserMapError {}

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum MapUsers {
    Uid(UntypedIdRange),
    Namespace(PathBuf),
}

impl From<UntypedIdRange> for MapUsers {
    fn from(value: UntypedIdRange) -> Self {
        Self::Uid(value)
    }
}

impl From<PathBuf> for MapUsers {
    fn from(value: PathBuf) -> Self {
        Self::Namespace(value)
    }
}

impl FromStr for MapUsers {
    type Err = ParseUserMapError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match UntypedIdRange::from_str(s) {
            Ok(range) => Ok(Self::from(range)),
            Err(_) => match PathBuf::from_str(s) {
                Ok(path) => Ok(Self::from(path)),
                Err(_) => Err(ParseUserMapError),
            },
        }
    }
}

impl Display for MapUsers {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                MapUsers::Uid(range) => range.to_string(),
                MapUsers::Namespace(path) => path.to_string_lossy().to_string(),
            }
        )
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseGidRangeError;

impl Display for ParseGidRangeError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "could not parse gid map")
    }
}

impl Error for ParseGidRangeError {}

impl From<ParseUntypedIdRangeError> for ParseGidRangeError {
    fn from(_value: ParseUntypedIdRangeError) -> Self {
        Self
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct GidRange(UntypedIdRange);

impl FromStr for GidRange {
    type Err = ParseGidRangeError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(UntypedIdRange::from_str(s).map(Self)?)
    }
}

impl Display for GidRange {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}

impl From<UntypedIdRange> for GidRange {
    fn from(value: UntypedIdRange) -> Self {
        Self(value)
    }
}

#[derive(Debug)]
pub struct Options<T>(Vec<T>);

impl<T> Default for Options<T> {
    fn default() -> Self {
        Self(Vec::new())
    }
}

impl<T> AsRef<Vec<T>> for Options<T> {
    fn as_ref(&self) -> &Vec<T> {
        &self.0
    }
}

impl<T> Deref for Options<T> {
    type Target = Vec<T>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl<T> DerefMut for Options<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.0
    }
}

impl<T: Clone> Clone for Options<T> {
    fn clone(&self) -> Self {
        Self(self.0.clone())
    }
}

impl<T: PartialEq> PartialEq for Options<T> {
    fn eq(&self, other: &Self) -> bool {
        self.0.eq(&other.0)
    }
}

impl<T: Eq> Eq for Options<T> {}

impl<T: FromStr> FromStr for Options<T> {
    type Err = T::Err;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        s.split(',')
            .map(T::from_str)
            .process_results(|iter| Options::from(iter.collect::<Vec<_>>()))
    }
}

impl<T: Display> Display for Options<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut iter = self.0.iter();
        if let Some(head) = iter.next() {
            write!(f, "{head}")?;

            for item in iter {
                write!(f, ",{item}")?;
            }
        }

        Ok(())
    }
}

impl<T> FromIterator<T> for Options<T> {
    fn from_iter<I: IntoIterator<Item = T>>(iter: I) -> Self {
        Self(iter.into_iter().collect())
    }
}

impl<T> From<Vec<T>> for Options<T> {
    fn from(value: Vec<T>) -> Self {
        Self(value)
    }
}

impl<T> From<T> for Options<T> {
    fn from(value: T) -> Self {
        Self(vec![value])
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct Sources(Options<OptionsSource>);

impl FromStr for Sources {
    type Err = ParseOptionsSourceError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(Self(Options::from_str(s)?))
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseUncheckedOptionsError;

impl Display for ParseUncheckedOptionsError {
    fn fmt(&self, _f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        unreachable!()
    }
}

impl Error for ParseUncheckedOptionsError {}

impl From<Infallible> for ParseUncheckedOptionsError {
    fn from(_value: Infallible) -> Self {
        Self
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct UncheckedOptions(Options<DisplayString>);

impl FromStr for UncheckedOptions {
    type Err = ParseUncheckedOptionsError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(Self(Options::<DisplayString>::from_str(s)?))
    }
}

impl Display for UncheckedOptions {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}

impl From<String> for UncheckedOptions {
    fn from(value: String) -> Self {
        Self::from_str(&value).unwrap()
    }
}

impl<T: AsRef<OsStr>> FromIterator<T> for UncheckedOptions {
    fn from_iter<I: IntoIterator<Item = T>>(iter: I) -> Self {
        Self(Options::from_iter(iter.into_iter().map(Into::into)))
    }
}

impl From<UncheckedOptions> for Options<DisplayString> {
    fn from(value: UncheckedOptions) -> Self {
        value.0
    }
}

impl AsRef<Options<DisplayString>> for UncheckedOptions {
    fn as_ref(&self) -> &Options<DisplayString> {
        &self.0
    }
}

impl From<Options<DisplayString>> for UncheckedOptions {
    fn from(value: Options<DisplayString>) -> Self {
        Self(value)
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum ParseMountOptionError {
    DeviceId(ParseDeviceIdTypeError),
    MapUsers(ParseUserMapError),
    MapGroups(ParseGidRangeError),
    OptionsMode(ParseOptionsModeError),
    OptionsSource(ParseOptionsSourceError),
    Operation(ParseMountOperationError),
    Subtree(ParseSubtreeError),
    UnknownOption(DisplayString),
}

impl Display for ParseMountOptionError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::DeviceId(e) => write!(f, "{e}"),
            Self::MapUsers(e) => write!(f, "{e}"),
            Self::MapGroups(e) => write!(f, "{e}"),
            Self::OptionsMode(e) => write!(f, "{e}"),
            Self::OptionsSource(e) => write!(f, "{e}"),
            Self::Operation(e) => write!(f, "{e}"),
            Self::Subtree(e) => write!(f, "{e}"),
            Self::UnknownOption(opt) => write!(f, "unknown option: {opt}"),
        }
    }
}

impl Error for ParseMountOptionError {}

impl From<ParseDeviceIdTypeError> for ParseMountOptionError {
    fn from(value: ParseDeviceIdTypeError) -> Self {
        Self::DeviceId(value)
    }
}

impl From<ParseUserMapError> for ParseMountOptionError {
    fn from(value: ParseUserMapError) -> Self {
        Self::MapUsers(value)
    }
}

impl From<ParseGidRangeError> for ParseMountOptionError {
    fn from(value: ParseGidRangeError) -> Self {
        Self::MapGroups(value)
    }
}

impl From<ParseOptionsModeError> for ParseMountOptionError {
    fn from(value: ParseOptionsModeError) -> Self {
        Self::OptionsMode(value)
    }
}

impl From<ParseUncheckedOptionsError> for ParseMountOptionError {
    fn from(_value: ParseUncheckedOptionsError) -> Self {
        unreachable!()
    }
}

impl From<ParseOptionsSourceError> for ParseMountOptionError {
    fn from(value: ParseOptionsSourceError) -> Self {
        Self::OptionsSource(value)
    }
}

impl From<ParseMountOperationError> for ParseMountOptionError {
    fn from(value: ParseMountOperationError) -> Self {
        Self::Operation(value)
    }
}

impl From<ParseSubtreeError> for ParseMountOptionError {
    fn from(value: ParseSubtreeError) -> Self {
        Self::Subtree(value)
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum MountOption {
    Label(DeviceId),
    FsType(UncheckedOptions),
    All,
    NoCanonicalize,
    Fake,
    Fork,
    Fstab(OsString),
    InternalOnly,
    ShowLabels,
    MapGroups(GidRange),
    MapUsers(MapUsers),
    MakeDir(Option<Permissions>),
    NoMtab,
    OptionsMode(OptionsMode),
    OptionsSource(Options<OptionsSource>),
    OptionsSourceForce,
    OnlyOnce,
    Options(UncheckedOptions),
    TestOptions(UncheckedOptions),
    Source(DeviceId),
    Target(PathBuf),
    TargetPrefix(PathBuf),
    Verbose,
    Access(Access),
    Namespace(OsString),
    Operation(MountOperation),
    Subtree(Subtree),
}

impl FromStr for MountOption {
    type Err = ParseMountOptionError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let s = s.trim().trim_start_matches('-');

        let option = match s.split_once(|delim| delim == ' ' || delim == '=') {
            Some((option, value)) => (option, Some(value)),
            None => (s, None),
        };

        match option {
            ("L" | "label", Some(id)) => Ok(Self::Label(DeviceId::from_str(id)?)),
            ("t" | "types", Some(types)) => Ok(Self::FsType(UncheckedOptions::from_str(types)?)),
            ("a" | "all", None) => Ok(Self::All),
            ("c" | "no-canonicalize", None) => Ok(Self::NoCanonicalize),
            ("f" | "fake", None) => Ok(Self::Fake),
            ("F" | "fork", None) => Ok(Self::Fork),
            ("T" | "fstab", Some(path)) => Ok(Self::Fstab(path.into())),
            ("i" | "internal-only", None) => Ok(Self::InternalOnly),
            ("l" | "show-labels", None) => Ok(Self::ShowLabels),
            ("map-groups", Some(map)) => Ok(Self::MapGroups(GidRange::from_str(map)?)),
            ("map-users", Some(map)) => Ok(Self::MapUsers(MapUsers::from_str(map)?)),
            ("m" | "mkdir", arg) => Ok(Self::MakeDir(
                arg.and_then(|arg| Permissions::from_str(arg).ok()),
            )),
            ("n" | "no-mtab", None) => Ok(Self::NoMtab),
            ("options-mode", Some(mode)) => Ok(Self::OptionsMode(OptionsMode::from_str(mode)?)),
            ("options-source", Some(source)) => Ok(Self::OptionsSource(Options::from_str(source)?)),
            ("options-source-force", None) => Ok(Self::OptionsSourceForce),
            ("onlyonce", None) => Ok(Self::OnlyOnce),
            ("o" | "options", Some(options)) => {
                Ok(Self::Options(UncheckedOptions::from_str(options)?))
            }
            ("O" | "test-opts", Some(options)) => {
                Ok(Self::TestOptions(UncheckedOptions::from_str(options)?))
            }
            ("r" | "read-only", None) => Ok(Self::Access(Access::ReadOnly)),
            ("source", Some(src)) => Ok(Self::Source(DeviceId::from_str(src)?)),
            ("target", Some(target)) => Ok(Self::Target(target.into())),
            ("target-prefix", Some(path)) => Ok(Self::TargetPrefix(path.into())),
            ("w" | "rw" | "read-write", None) => Ok(Self::Access(Access::ReadWrite)),
            ("N" | "namespace", Some(ns)) => Ok(Self::Namespace(ns.into())),
            ("U" | "uuid", Some(uuid)) => Ok(Self::Label(DeviceId::Uuid(uuid.into()))),
            ("B" | "bind", None) => Ok(Self::Operation(MountOperation::Bind)),
            ("M" | "move", None) => Ok(Self::Operation(MountOperation::Move)),
            ("R" | "rbind", None) => Ok(Self::Operation(MountOperation::RecursiveBind)),
            (op, None) if op.starts_with("make-") => Ok(Self::Subtree(Subtree::from_str(
                op.strip_prefix("make-").unwrap(),
            )?)),
            (opt, _) => Err(ParseMountOptionError::UnknownOption(opt.into())),
        }
    }
}

impl Display for MountOption {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Self::Label(id) => format!("--label {id}"),
                Self::FsType(t) => format!("--types {t}"),
                Self::All => "--all".to_string(),
                Self::NoCanonicalize => "--no-canonicalize".to_string(),
                Self::Fake => "--fake".to_string(),
                Self::Fork => "--fork".to_string(),
                Self::Fstab(path) => format!("--fstab {}", path.to_string_lossy()),
                Self::InternalOnly => "--internal-only".to_string(),
                Self::ShowLabels => "--show-labels".to_string(),
                Self::MapGroups(range) => format!("--map-groups {range}"),
                Self::MapUsers(map) => format!("--map-users {map}"),
                Self::MakeDir(mode) => format!(
                    "--mkdir {}",
                    mode.clone()
                        .map(|m| m.to_string())
                        .unwrap_or_else(String::new)
                ),
                Self::NoMtab => "--no-mtab".to_string(),
                Self::OptionsMode(mode) => format!("--options-mode {mode}"),
                Self::OptionsSource(source) => format!("--options-source {source}"),
                Self::OptionsSourceForce => "--options-source-force".to_string(),
                Self::OnlyOnce => "--onlyonce".to_string(),
                Self::Options(opts) => format!("--options {opts}"),
                Self::TestOptions(opts) => format!("--test-opts {opts}"),
                Self::Source(src) => format!("--source {src}"),
                Self::Target(path) => format!("--target {}", path.to_string_lossy()),
                Self::TargetPrefix(path) => format!("--target-prefix {}", path.to_string_lossy()),
                Self::Verbose => "--verbose".to_string(),
                Self::Access(access) => format!("--{access}"),
                Self::Namespace(ns) => format!("--namespace {}", ns.to_string_lossy()),
                Self::Operation(op) => format!("--{op}"),
                Self::Subtree(op) => format!("--make-{op}"),
            }
        )
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseSubtreeError(DisplayString);

impl Display for ParseSubtreeError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "received invalid subtree: {}", self.0)
    }
}

impl Error for ParseSubtreeError {}

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum Subtree {
    Shared,
    Replica,
    Private,
    Unbindable,
    Recursive(Box<Subtree>),
}

impl Subtree {
    pub fn recursive(self) -> Self {
        Self::Recursive(Box::new(self))
    }
}

impl FromStr for Subtree {
    type Err = ParseSubtreeError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "shared" => Ok(Self::Shared),
            "slave" | "replica" => Ok(Self::Replica),
            "private" => Ok(Self::Private),
            "unbindable" => Ok(Self::Unbindable),

            "rshared" => Ok(Self::recursive(Self::Shared)),
            "rslave" | "rreplica" => Ok(Self::recursive(Self::Replica)),
            "rprivate" => Ok(Self::recursive(Self::Private)),
            "runbindable" => Ok(Self::recursive(Self::Unbindable)),

            _ => Err(ParseSubtreeError(s.into())),
        }
    }
}

impl Display for Subtree {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Subtree::Shared => write!(f, "shared"),
            Subtree::Replica => write!(f, "slave"),
            Subtree::Private => write!(f, "private"),
            Subtree::Unbindable => write!(f, "unbindable"),
            Subtree::Recursive(op) => write!(f, "r{op}"),
        }
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseMountOperationError(DisplayString);

impl Display for ParseMountOperationError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "could not parse operation: {}", self.0)
    }
}

impl Error for ParseMountOperationError {}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum MountOperation {
    Bind,
    RecursiveBind,
    Move,
}

impl FromStr for MountOperation {
    type Err = ParseMountOperationError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "bind" => Ok(Self::Bind),
            "rbind" => Ok(Self::RecursiveBind),
            "move" => Ok(Self::Move),
            v => Err(ParseMountOperationError(v.into())),
        }
    }
}

impl Display for MountOperation {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                MountOperation::Bind => "bind",
                MountOperation::RecursiveBind => "rbind",
                MountOperation::Move => "move",
            }
        )
    }
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct Mount {
    device_id: DeviceId,
    mountpoint: PathBuf,
    options: Options<MountOption>,
}

impl Mount {
    pub fn new<T, I>(source: impl Into<DeviceId>, target: impl Into<PathBuf>, options: I) -> Self
    where
        T: Into<MountOption>,
        I: IntoIterator<Item = T>,
    {
        Self {
            device_id: source.into(),
            mountpoint: target.into(),
            options: Options::from_iter(options.into_iter().map(Into::into)),
        }
    }
}

impl FileSystem for Mount {
    fn mount(&mut self) -> std::io::Result<()> {
        mount(
            self.device_id,
            &self.mountpoint,
            self.options.map(Into::into),
        )?;
        Ok(())
    }

    fn unmount(&mut self) -> std::io::Result<()> {
        unmount(&self.mountpoint)?;
        Ok(())
    }
}

pub fn mount<T: AsRef<MountOption>, I: IntoIterator<Item = T>>(
    source: impl Into<DeviceId>,
    target: impl Into<PathBuf>,
    options: I,
) -> std::io::Result<std::process::Output> {
    Command::new("mount")
        .args(options.into_iter().map(|s| s.as_ref().to_string()))
        .arg(source.into().to_string())
        .arg(target.into())
        .output()
}

pub fn unmount(target: impl Into<PathBuf>) -> std::io::Result<std::process::Output> {
    Command::new("umount").arg(target.into()).output()
}

#[cfg(test)]
mod tests {
    use std::{borrow::Cow, path::PathBuf, str::FromStr};

    use enumflags2::BitFlags;

    use crate::fs::{
        id_mapping::UntypedIdRange,
        mount::{ParseDeviceIdTypeError, ParseOptionsSourceError},
        permission::Permissions,
    };

    use super::{
        Access, DeviceId, KeyValuePair, MapUsers, MountOperation, MountOption, Options,
        OptionsMode, OptionsSource, ParseMountOptionError, Subtree, UncheckedOptions,
    };

    #[test]
    fn key_value() {
        let kv = KeyValuePair::from_str("test=value").unwrap();
        let k = KeyValuePair::from_str("keyword").unwrap();

        assert_eq!(kv, KeyValuePair::new("test", Some("value")));
        assert_eq!(k, KeyValuePair::new("keyword", None::<String>));

        assert_eq!(kv.to_string(), "test=value");
        assert_eq!(k.to_string(), "keyword");
    }

    #[test]
    fn access() {
        let read = Access::from_str("read-only").unwrap();

        let read_write = Access::from_str("read-write").unwrap();

        assert_eq!(read, Access::ReadOnly);
        assert_eq!(read_write, Access::ReadWrite);

        assert_eq!(read.to_string(), "read-only");
        assert_eq!(read_write.to_string(), "read-write");
    }

    #[test]
    fn device_id() {
        let label = DeviceId::from_str("LABEL=test").unwrap();
        let uuid = DeviceId::from_str("UUID=1234").unwrap();
        let pl = DeviceId::from_str("PARTLABEL=test").unwrap();
        let pu = DeviceId::from_str("PARTUUID=98776543").unwrap();
        let id = DeviceId::from_str("ID=awawa").unwrap();
        let other = DeviceId::from_str("awawa").unwrap();
        let should_err = DeviceId::from_str("GARBAGE=trash");

        assert_eq!(label, DeviceId::Label("test".into()));
        assert_eq!(uuid, DeviceId::Uuid("1234".into()));
        assert_eq!(pl, DeviceId::PartLabel("test".into()));
        assert_eq!(pu, DeviceId::PartUuid("98776543".into()));
        assert_eq!(id, DeviceId::Id("awawa".into()));
        assert_eq!(other, DeviceId::Other("awawa".into()));
        assert_eq!(
            should_err,
            Err(ParseDeviceIdTypeError("GARBAGE".to_string()))
        );

        assert_eq!(label.to_string(), "LABEL=test");
        assert_eq!(uuid.to_string(), "UUID=1234");
        assert_eq!(pl.to_string(), "PARTLABEL=test");
        assert_eq!(pu.to_string(), "PARTUUID=98776543");
        assert_eq!(id.to_string(), "ID=awawa");
        assert_eq!(other.to_string(), "awawa");
    }

    #[test]
    fn options_mode() {
        let ignore = OptionsMode::from_str("ignore").unwrap();
        let append = OptionsMode::from_str("append").unwrap();
        let prepend = OptionsMode::from_str("prepend").unwrap();
        let replace = OptionsMode::from_str("replace").unwrap();

        assert_eq!(ignore, OptionsMode::Ignore);
        assert_eq!(append, OptionsMode::Append);
        assert_eq!(prepend, OptionsMode::Prepend);
        assert_eq!(replace, OptionsMode::Replace);

        assert_eq!(ignore.to_string(), "ignore");
        assert_eq!(append.to_string(), "append");
        assert_eq!(prepend.to_string(), "prepend");
        assert_eq!(replace.to_string(), "replace");
    }

    #[test]
    fn options_source() {
        let mtab = OptionsSource::from_str("mtab").unwrap();
        let fstab = OptionsSource::from_str("fstab").unwrap();
        let disable = OptionsSource::from_str("disable").unwrap();
        let should_err = OptionsSource::from_str("awawa");

        assert_eq!(mtab, OptionsSource::Mtab);
        assert_eq!(fstab, OptionsSource::Fstab);
        assert_eq!(disable, OptionsSource::Disable);
        assert_eq!(
            should_err,
            Err(ParseOptionsSourceError("awawa".to_string()))
        );

        assert_eq!(mtab.to_string(), "mtab");
        assert_eq!(fstab.to_string(), "fstab");
        assert_eq!(disable.to_string(), "disable");
    }

    #[test]
    fn map_users() {
        let uid_map = MapUsers::from_str("1000:1001:1").unwrap();
        let ns_map = MapUsers::from_str("/proc/1000/ns/user").unwrap();

        assert_eq!(uid_map, MapUsers::Uid(UntypedIdRange(1000, 1001, 1)));
        assert_eq!(
            ns_map,
            MapUsers::Namespace(PathBuf::from_str("/proc/1000/ns/user").unwrap())
        );

        assert_eq!(uid_map.to_string(), "1000:1001:1");
        assert_eq!(ns_map.to_string(), "/proc/1000/ns/user");
    }

    #[test]
    fn options() {
        let opts = Options::<KeyValuePair>::from_str("key=value,keyword").unwrap();

        assert_eq!(
            opts,
            Options(vec![
                KeyValuePair::new("key", Some("value")),
                KeyValuePair::new("keyword", None::<&str>)
            ])
        );

        assert_eq!(opts.to_string(), "key=value,keyword");
    }

    #[test]
    fn operations() {
        let bind = MountOperation::from_str("bind").unwrap();
        let mov = MountOperation::from_str("move").unwrap();
        let rbind = MountOperation::from_str("rbind").unwrap();

        assert_eq!(bind, MountOperation::Bind);
        assert_eq!(mov, MountOperation::Move);
        assert_eq!(rbind, MountOperation::RecursiveBind);

        assert_eq!(bind.to_string(), "bind");
        assert_eq!(mov.to_string(), "move");
        assert_eq!(rbind.to_string(), "rbind");
    }

    #[test]
    fn subtree() {
        let shared = Subtree::from_str("shared").unwrap();
        let rshared = Subtree::from_str("rshared").unwrap();
        let private = Subtree::from_str("private").unwrap();
        let rprivate = Subtree::from_str("rprivate").unwrap();
        let replica = Subtree::from_str("replica").unwrap();
        let rreplica = Subtree::from_str("rreplica").unwrap();
        let unbind = Subtree::from_str("unbindable").unwrap();
        let runbind = Subtree::from_str("runbindable").unwrap();

        assert_eq!(shared, Subtree::Shared);
        assert_eq!(rshared, Subtree::recursive(Subtree::Shared));
        assert_eq!(private, Subtree::Private);
        assert_eq!(rprivate, Subtree::recursive(Subtree::Private));
        assert_eq!(replica, Subtree::Replica);
        assert_eq!(rreplica, Subtree::recursive(Subtree::Replica));
        assert_eq!(unbind, Subtree::Unbindable);
        assert_eq!(runbind, Subtree::recursive(Subtree::Unbindable));

        assert_eq!(shared.to_string(), "shared");
        assert_eq!(rshared.to_string(), "rshared");
        assert_eq!(private.to_string(), "private");
        assert_eq!(rprivate.to_string(), "rprivate");
        assert_eq!(replica.to_string(), "slave");
        assert_eq!(rreplica.to_string(), "rslave");
        assert_eq!(unbind.to_string(), "unbindable");
        assert_eq!(runbind.to_string(), "runbindable");
    }

    fn try_opts<'a, 'b>(
        cmds: &[&'a str],
        args: impl Into<Option<&'b str>>,
    ) -> Result<MountOption, ParseMountOptionError>
    where
        'a: 'b,
    {
        let args = args.into();
        let to_args = |c: &'a str| -> Cow<'b, str> {
            match args {
                Some(x) => format!("{c} {x}").into(),
                None => c.into(),
            }
        };

        cmds.iter().map(|s| to_args(s)).fold(
            Err(ParseMountOptionError::UnknownOption("".into())),
            |_, c| MountOption::from_str(&c),
        )
    }

    struct Opts<'a, 'b> {
        options: Cow<'a, [&'a str]>,
        args: Option<&'b str>,
    }

    impl<'a, 'b> Opts<'a, 'b> {
        fn assert(&self, expected: MountOption) {
            let opt = try_opts(&self.options, self.args).unwrap();
            assert_eq!(opt, expected);
        }
    }

    impl<'a, 'b> From<&'a str> for Opts<'a, 'b> {
        fn from(value: &'a str) -> Self {
            Self {
                options: vec![value].into(),
                args: None,
            }
        }
    }

    impl<'a, 'b> From<(&'a str, &'b str)> for Opts<'a, 'b> {
        fn from(value: (&'a str, &'b str)) -> Self {
            Self {
                options: vec![value.0].into(),
                args: Some(value.1),
            }
        }
    }

    impl<'a, 'b> From<&'a [&'a str]> for Opts<'a, 'b> {
        fn from(value: &'a [&'a str]) -> Self {
            Self {
                options: value.into(),
                args: None,
            }
        }
    }

    impl<'a, 'b> From<(&'a [&str], &'b str)> for Opts<'a, 'b> {
        fn from(value: (&'a [&str], &'b str)) -> Self {
            Self {
                options: value.0.into(),
                args: Some(value.1),
            }
        }
    }

    #[test]
    fn mount_options() {
        Opts::from((["-L", "--label"].as_slice(), "LABEL=label"))
            .assert(MountOption::Label(DeviceId::Label("label".to_string())));

        Opts::from((["-t", "--types"].as_slice(), "nomsdos,smbs")).assert(MountOption::FsType(
            UncheckedOptions::from_iter(vec!["nomsdos", "smbs"]),
        ));

        Opts::from("-a").assert(MountOption::All);
        Opts::from(["-c", "--no-canonicalize"].as_slice()).assert(MountOption::NoCanonicalize);
        Opts::from(["-f", "--fake"].as_slice()).assert(MountOption::Fake);
        Opts::from(["-F", "--fork"].as_slice()).assert(MountOption::Fork);

        Opts::from((["-T", "--fstab"].as_slice(), "/etc/fstab"))
            .assert(MountOption::Fstab("/etc/fstab".into()));

        Opts::from(["-i", "--internal-only"].as_slice()).assert(MountOption::InternalOnly);
        Opts::from(["-l", "--show-labels"].as_slice()).assert(MountOption::ShowLabels);

        Opts::from(("--map-groups", "1000:1001:1"))
            .assert(MountOption::MapGroups(UntypedIdRange(1000, 1001, 1).into()));

        Opts::from(("--map-users", "1000:1001:1")).assert(MountOption::MapUsers(MapUsers::Uid(
            UntypedIdRange(1000, 1001, 1),
        )));

        Opts::from(("--map-users", "/proc/1000/ns/user")).assert(MountOption::MapUsers(
            MapUsers::Namespace("/proc/1000/ns/user".into()),
        ));

        Opts::from(["-m", "--mkdir"].as_slice()).assert(MountOption::MakeDir(None));

        Opts::from((["-m", "--mkdir"].as_slice(), "7777")).assert(MountOption::MakeDir(Some(
            Permissions::Octal(BitFlags::all().into()),
        )));

        Opts::from(["-n", "--no-mtab"].as_slice()).assert(MountOption::NoMtab);

        Opts::from(("--options-mode", "ignore"))
            .assert(MountOption::OptionsMode(OptionsMode::Ignore));

        Opts::from(("--options-source", "fstab")).assert(MountOption::OptionsSource(
            vec![OptionsSource::Fstab].into(),
        ));

        Opts::from("--options-source-force").assert(MountOption::OptionsSourceForce);
        Opts::from("--onlyonce").assert(MountOption::OnlyOnce);

        let opts = UncheckedOptions::from_iter(vec!["kbity", "dogy", "aaaaaa=awawa"]);

        Opts::from((["-o", "--options"].as_slice(), "kbity,dogy,aaaaaa=awawa"))
            .assert(MountOption::Options(opts.clone()));

        Opts::from((["-O", "--test-opts"].as_slice(), "kbity,dogy,aaaaaa=awawa"))
            .assert(MountOption::TestOptions(opts));

        Opts::from(["-r", "--read-only"].as_slice()).assert(MountOption::Access(Access::ReadOnly));

        Opts::from(("--source", "LABEL=label"))
            .assert(MountOption::Source(DeviceId::Label("label".into())));

        Opts::from(("--target", "/mnt")).assert(MountOption::Target("/mnt".into()));
        Opts::from(("--target-prefix", "/pfx")).assert(MountOption::TargetPrefix("/pfx".into()));

        Opts::from(["-w", "--rw", "--read-write"].as_slice())
            .assert(MountOption::Access(Access::ReadWrite));

        Opts::from((["-N", "--namespace"].as_slice(), "1000"))
            .assert(MountOption::Namespace("1000".into()));

        Opts::from((["-U", "--uuid"].as_slice(), "1234"))
            .assert(MountOption::Label(DeviceId::Uuid("1234".into())));

        Opts::from(["-B", "--bind"].as_slice())
            .assert(MountOption::Operation(MountOperation::Bind));
        Opts::from(["-M", "--move"].as_slice())
            .assert(MountOption::Operation(MountOperation::Move));
        Opts::from(["-R", "--rbind"].as_slice())
            .assert(MountOption::Operation(MountOperation::RecursiveBind));

        Opts::from("--make-shared").assert(MountOption::Subtree(Subtree::Shared));
        Opts::from("--make-rshared")
            .assert(MountOption::Subtree(Subtree::recursive(Subtree::Shared)));

        Opts::from("--make-private").assert(MountOption::Subtree(Subtree::Private));
        Opts::from("--make-rprivate")
            .assert(MountOption::Subtree(Subtree::recursive(Subtree::Private)));

        Opts::from("--make-slave").assert(MountOption::Subtree(Subtree::Replica));
        Opts::from("--make-rslave")
            .assert(MountOption::Subtree(Subtree::recursive(Subtree::Replica)));

        Opts::from("--make-unbindable").assert(MountOption::Subtree(Subtree::Unbindable));
        Opts::from("--make-runbindable").assert(MountOption::Subtree(Subtree::recursive(
            Subtree::Unbindable,
        )));
    }
}
