use std::ffi::OsString;
use std::fmt::Display;
use std::path::PathBuf;
use std::str::FromStr;

use char_enum::FromStrError;

use crate::fs::mount::{mount, unmount, MountOption, UncheckedOptions};
use crate::fs::FileSystem;
use crate::macros::{impl_deref, impl_deref_mut, unwrap_enum};

use super::{LowerDirs, Stack, StackFs};

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum Switch {
    On,
    Off,
}

impl FromStr for Switch {
    type Err = FromStrError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.to_lowercase().as_str() {
            "on" => Ok(Self::On),
            "off" => Ok(Self::Off),
            s => Err(FromStrError(s.to_string())),
        }
    }
}

impl Display for Switch {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Switch::On => "on",
                Switch::Off => "off",
            }
        )
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum RedirectDir {
    On,
    Follow,
    NoFollow,
    Off,
}

impl FromStr for RedirectDir {
    type Err = FromStrError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.to_lowercase().as_str() {
            "on" => Ok(Self::On),
            "follow" => Ok(Self::Follow),
            "nofollow" => Ok(Self::NoFollow),
            "off" => Ok(Self::Off),
            s => Err(FromStrError(s.to_string())),
        }
    }
}

impl Display for RedirectDir {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                RedirectDir::On => "on",
                RedirectDir::Follow => "follow",
                RedirectDir::NoFollow => "nofollow",
                RedirectDir::Off => "off",
            }
        )
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum Uuid {
    On,
    Auto,
    Null,
    Off,
}

impl FromStr for Uuid {
    type Err = FromStrError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.to_lowercase().as_str() {
            "on" => Ok(Self::On),
            "auto" => Ok(Self::Auto),
            "null" => Ok(Self::Null),
            "off" => Ok(Self::Off),
            s => Err(FromStrError(s.to_string())),
        }
    }
}

impl Display for Uuid {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Self::On => "on",
                Self::Auto => "auto",
                Self::Null => "null",
                Self::Off => "off",
            }
        )
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum Verity {
    On,
    Require,
    Off,
}

impl FromStr for Verity {
    type Err = FromStrError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.to_lowercase().as_str() {
            "on" => Ok(Self::On),
            "require" => Ok(Self::Require),
            "off" => Ok(Self::Off),
            s => Err(FromStrError(s.to_string())),
        }
    }
}

impl Display for Verity {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Self::On => "on",
                Self::Require => "require",
                Self::Off => "off",
            }
        )
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum Xino {
    On,
    Auto,
    Off,
}

impl FromStr for Xino {
    type Err = FromStrError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.to_lowercase().as_str() {
            "on" => Ok(Self::On),
            "auto" => Ok(Self::Auto),
            "off" => Ok(Self::Off),
            s => Err(FromStrError(s.to_string())),
        }
    }
}

impl Display for Xino {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Self::On => "on",
                Self::Auto => "Auto",
                Self::Off => "off",
            }
        )
    }
}

#[derive(Debug)]
pub enum OverlayOption {
    Index(Switch),
    LowerDir(LowerDirs),
    Metacopy(Switch),
    NfsExport(Switch),
    RedirectDir(RedirectDir),
    UpperDir(PathBuf),
    UserXAttr,
    Uuid(Uuid),
    Verity(Verity),
    Volatile,
    WorkDir(PathBuf),
    Xino(Xino),
}

impl FromStr for OverlayOption {
    type Err = FromStrError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let s = s.trim();
        let arg = match s.split_once('=') {
            Some((k, v)) => (k, Some(v)),
            None => (s, None),
        };

        match arg {
            ("index", Some(enabled)) => Ok(Self::Index(Switch::from_str(enabled)?)),
            ("lowerdir", Some(dirs)) => Ok(Self::LowerDir(LowerDirs::from_str(dirs).unwrap())),
            ("metacopy", Some(enabled)) => Ok(Self::Metacopy(Switch::from_str(enabled)?)),
            ("nfs_export", Some(enabled)) => Ok(Self::NfsExport(Switch::from_str(enabled)?)),
            ("redirect_dir", Some(dir)) => Ok(Self::RedirectDir(RedirectDir::from_str(dir)?)),
            ("upperdir", Some(dir)) => Ok(Self::UpperDir(PathBuf::from_str(dir).unwrap())),
            ("userxattr", None) => Ok(Self::UserXAttr),
            ("uuid", Some(uuid)) => Ok(Self::Uuid(Uuid::from_str(uuid)?)),
            ("verity", Some(verity)) => Ok(Self::Verity(Verity::from_str(verity)?)),
            ("volatile", None) => Ok(Self::Volatile),
            ("workdir", Some(dir)) => Ok(Self::WorkDir(PathBuf::from_str(dir).unwrap())),
            ("xino", Some(xino)) => Ok(Self::Xino(Xino::from_str(xino)?)),
            (arg, _) => Err(FromStrError(arg.to_string())),
        }
    }
}

impl Display for OverlayOption {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Index(enabled) => write!(f, "index={enabled}"),
            Self::LowerDir(dirs) => write!(f, "lowerdir={dirs}"),
            Self::Metacopy(enabled) => write!(f, "metacopy={enabled}"),
            Self::NfsExport(enabled) => write!(f, "nfs_export={enabled}"),
            Self::RedirectDir(option) => write!(f, "redirect_dir={option}"),
            Self::UpperDir(dir) => write!(f, "upperdir={}", dir.to_string_lossy()),
            Self::UserXAttr => write!(f, "userxattr"),
            Self::Uuid(uuid) => write!(f, "uid={uuid}"),
            Self::Verity(verity) => write!(f, "verity={verity}"),
            Self::Volatile => write!(f, "volatile"),
            Self::WorkDir(dir) => write!(f, "workdir={}", dir.to_string_lossy()),
            Self::Xino(xino) => write!(f, "xino={xino}"),
        }
    }
}

#[derive(Debug)]
pub struct Overlay(StackFs<OverlayOption>);

impl_deref!(Overlay(StackFs<OverlayOption>));
impl_deref_mut!(Overlay(StackFs<OverlayOption>));

impl Stack for Overlay {
    fn lower_dirs(&self) -> impl Iterator<Item = &std::path::Path> {
        let result = self.find_option(|v| unwrap_enum!(v, OverlayOption::LowerDir));

        match result {
            Some(dirs) => dirs.into_iter(),
            None => Default::default(),
        }
    }

    fn upper_dir(&self) -> Option<&std::path::Path> {
        self.find_option(|v| unwrap_enum!(v, OverlayOption::UpperDir))
            .map(PathBuf::as_path)
    }

    fn work_dir(&self) -> Option<&std::path::Path> {
        self.find_option(|v| unwrap_enum!(v, OverlayOption::WorkDir))
            .map(PathBuf::as_path)
    }
}

impl FileSystem for Overlay {
    fn mount(&mut self) -> std::io::Result<()> {
        let opts = self.options.iter().map(ToString::to_string);
        let options = MountOption::Options(UncheckedOptions::from_iter(opts));
        mount("overlay", &self.mountpoint, options)?;
        Ok(())
    }

    fn unmount(&mut self) -> std::io::Result<()> {
        unmount(&self.mountpoint)?;
        Ok(())
    }
}
