macro_rules! impl_fromstr {
    ($name:ident($inner:ty), $err:ty) => {
        impl std::str::FromStr for $name {
            type Err = $err;

            fn from_str(s: &str) -> Result<Self, Self::Err> {
                Ok(Self(<$inner>::from_str(s)?))
            }
        }
    };
}

macro_rules! impl_display {
    ($name:ident) => {
        impl Display for $name {
            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                write!(f, "{}", self.0)
            }
        }
    };
}

macro_rules! impl_deref {
    ($name:ident($inner:ty)) => {
        impl std::ops::Deref for $name {
            type Target = $inner;

            fn deref(&self) -> &Self::Target {
                &self.0
            }
        }
    };
}

macro_rules! impl_deref_mut {
    ($name:ident($inner:ty)) => {
        impl std::ops::DerefMut for $name {
            fn deref_mut(&mut self) -> &mut Self::Target {
                &mut self.0
            }
        }
    };
}

macro_rules! impl_from_variants {
    ($enum_type:ident, $($variant:ident($inner_type:ty)),*) => {
        $(
            impl From<$inner_type> for $enum_type {
                fn from(inner: $inner_type) -> Self {
                    $enum_type::$variant(inner)
                }
            }
        )*
    };
}

macro_rules! impl_wrapper_err_for {
    ($name:ident($inner:ty), $err:literal) => {
        impl_wrapper_err!($name, $err);
        impl_from_ty!($name, $inner);
    };
}

macro_rules! impl_from_ty {
    ($name:ty, $from:ty) => {
        impl From<$from> for $name {
            fn from(value: $from) -> Self {
                Self(value)
            }
        }
    };
}

macro_rules! impl_wrapper_err {
    ($type:ty, $err:literal) => {
        impl Display for $type {
            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                write!(f, "{}: {}", self.0, $err)
            }
        }

        impl Error for $type {}
    };
}

macro_rules! unwrap_enum {
    ($value:ident, $enum:ident::$variant:ident) => {
        match $value {
            $enum::$variant(v) => Some(v),
            _ => None,
        }
    };
}

pub(crate) use impl_deref;
pub(crate) use impl_deref_mut;
pub(crate) use impl_display;
pub(crate) use impl_from_ty;
pub(crate) use impl_from_variants;
pub(crate) use impl_fromstr;
pub(crate) use impl_wrapper_err;
pub(crate) use impl_wrapper_err_for;
pub(crate) use unwrap_enum;
