use core::fmt::Debug;
use std::convert::Infallible;
use std::ffi::{OsStr, OsString};
use std::fmt::Display;
use std::path::Path;
use std::process::{Command as StdCommand, Output};

use crate::macros::{impl_deref, impl_deref_mut, impl_fromstr};

pub trait MapWhileRefExt<'a, I: 'a> {
    fn map_while_ref<F>(&'a mut self, f: F) -> MapWhileRef<'a, I, F>;
}

impl<'a, I: Iterator + 'a> MapWhileRefExt<'a, I> for I {
    fn map_while_ref<F>(&mut self, f: F) -> MapWhileRef<I, F> {
        MapWhileRef { iter: self, f }
    }
}

#[derive(Debug)]
pub struct MapWhileRef<'a, I: 'a, F> {
    iter: &'a mut I,
    f: F,
}

impl<I, F, U, E> Iterator for MapWhileRef<'_, I, F>
where
    I: Iterator + Clone,
    F: FnMut(&I::Item) -> Result<U, E>,
{
    type Item = U;

    fn next(&mut self) -> Option<Self::Item> {
        let old = self.iter.clone();
        match self.iter.next() {
            None => None,
            Some(elt) => match (self.f)(&elt) {
                Ok(val) => Some(val),
                Err(_) => {
                    *self.iter = old;
                    None
                }
            },
        }
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        (0, self.iter.size_hint().1)
    }
}

#[derive(Debug)]
pub enum CommandError {
    Fail(Output),
    Error(std::io::Error),
}

impl Display for CommandError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            CommandError::Fail(output) => match output.status.code() {
                Some(code) => {
                    write!(
                        f,
                        "exit {code}: {}",
                        String::from_utf8_lossy(&output.stderr)
                    )
                }
                None => write!(f, "process terminated by signal"),
            },
            CommandError::Error(error) => write!(f, "{error:?}"),
        }
    }
}

impl From<std::io::Error> for CommandError {
    fn from(value: std::io::Error) -> Self {
        Self::Error(value)
    }
}

impl From<CommandError> for std::io::Error {
    fn from(value: CommandError) -> Self {
        match value {
            CommandError::Fail(output) => {
                std::io::Error::other(String::from_utf8_lossy(&output.stderr))
            }

            CommandError::Error(error) => error,
        }
    }
}

#[derive(Debug)]
pub struct Command(StdCommand);

impl_deref!(Command(StdCommand));
impl_deref_mut!(Command(StdCommand));

impl Command {
    pub fn new(cmd: impl AsRef<OsStr>) -> Self {
        Self(StdCommand::new(cmd))
    }

    pub fn arg(&mut self, arg: impl AsRef<OsStr>) -> &mut Self {
        self.0.arg(arg);
        self
    }

    pub fn args<I, S>(&mut self, args: I) -> &mut Self
    where
        I: IntoIterator<Item = S>,
        S: AsRef<OsStr>,
    {
        self.0.args(args);
        self
    }

    pub fn env(&mut self, key: impl AsRef<OsStr>, value: impl AsRef<OsStr>) -> &mut Self {
        self.0.env(key, value);
        self
    }

    pub fn env_remove(&mut self, key: impl AsRef<OsStr>) -> &mut Self {
        self.0.env_remove(key);
        self
    }

    pub fn current_dir(&mut self, path: impl AsRef<Path>) -> &mut Self {
        self.0.current_dir(path);
        self
    }

    pub fn output(&mut self) -> Result<Output, CommandError> {
        let output = self.0.output()?;

        if output.status.success() {
            Ok(output)
        } else {
            Err(CommandError::Fail(output))
        }
    }
}

#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct DisplayString(OsString);

impl_deref!(DisplayString(OsString));
impl_deref_mut!(DisplayString(OsString));
impl_fromstr!(DisplayString(OsString), Infallible);

impl<T: AsRef<OsStr>> From<T> for DisplayString {
    fn from(value: T) -> Self {
        Self(value.as_ref().into())
    }
}

impl Display for DisplayString {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0.display())
    }
}
